import argparse
import ipaddress
import logging
import time
from dataclasses import dataclass
from typing import Tuple

from clippy.cookbook import Cookbook
from clippy.exceptions import AutomationError
from clippy.remote import RemoteCommandError, RemoteHosts

from rover.netbox import Netbox
from rover.os import DEFAULT_NEW_OS, PRIMARY_INTERFACE_NAME, DebianVersion, DiskPartitioningScheme
from rover.virt import GanetiCluster, Hypervisor, ProxmoxCluster, VirtCluster, VirtualMachine, find_hv_cluster

FIND_NETWORK_DATA_QUERY = """
query get_prefixes($name: String!) {
    prefix_list(filters: {description: {exact: $name}}) {
        id
        description
        prefix

        family {
            value
            label
        }
        
        vlan {
            id
            name
        }
    }
}
"""


@dataclass
class NetworkData:
    netbox_name: str
    vlan_id: int | None
    subnet_v4_id: int
    subnet_v6: ipaddress.IPv6Network


class CreateVm(Cookbook):
    """Creates a new virtual machine."""

    def argument_parser(self) -> argparse.ArgumentParser:
        parser = super().argument_parser()
        parser.add_argument("fqdn", help="full host name of the new virtual machine")
        parser.add_argument("--virt", help="virtualization host fqdn", type=str, required=True)
        parser.add_argument(
            "--os",
            help="operating system (Debian) version codename to use",
            type=DebianVersion,
            choices=list(DebianVersion),
            default=DEFAULT_NEW_OS,
        )
        parser.add_argument(
            "--disk-layout",
            help="disk provisioning scheme",
            type=DiskPartitioningScheme,
            choices=list(DiskPartitioningScheme),
            default=DiskPartitioningScheme.FLAT,
        )
        parser.add_argument(
            "--disk-space",
            help="amount of disk space in GiB to allocate",
            type=int,
            default=8,
        )
        parser.add_argument("--cpu", help="number of vCPUs to allocate", type=int, default=1)
        parser.add_argument("--mem", help="number of GiBs of RAM to allocate", type=float, default=1)
        parser.add_argument("--network", help="name of the network to connect to", type=str, required=True)

        return parser

    def _get_virt_details(self, netbox: Netbox) -> tuple[VirtCluster, Hypervisor]:
        self.logger.debug("Querying Netbox for hypervisor %s", self.args.virt)

    def _get_network_data(self, netbox: Netbox) -> NetworkData:
        self.logger.debug("Querying Netbox for network %s", self.args.network)
        nb_results = netbox.query(gql=FIND_NETWORK_DATA_QUERY, op="get_prefixes", variables={"name": self.args.network})

        name = None
        vlan_id = None
        subnet_v4 = None
        subnet_v4_id = None
        subnet_v6 = None
        for prefix in nb_results["prefix_list"]:
            if name and name != prefix["description"]:
                raise AutomationError(f"Found conflicting prefix names: '{name}' != '{prefix['description']}'")
            name = prefix["description"]

            if prefix["family"]["value"] == 6:
                subnet_v6 = ipaddress.IPv6Network(prefix["prefix"])
            else:
                subnet_v4 = ipaddress.IPv4Network(prefix["prefix"])
                subnet_v4_id = int(prefix["id"])

            this_prefix_vlan_id = (prefix.get("vlan") or {}).get("id")
            if this_prefix_vlan_id:
                if vlan_id and int(this_prefix_vlan_id) != vlan_id:
                    raise AutomationError(f"Found conflicting VLANs: {vlan_id} != {this_prefix_vlan_id}")
                if not vlan_id:
                    vlan_id = int(this_prefix_vlan_id)

        self.logger.info("Found network %s: %s / %s (vlan %s)", name, subnet_v4, subnet_v6, vlan_id)

        return NetworkData(
            netbox_name=name,
            vlan_id=vlan_id,
            subnet_v4_id=subnet_v4_id,
            subnet_v6=subnet_v6,
        )

    def _allocate_addresses(self, netbox: Netbox, network_data: NetworkData) -> Tuple[str, str]:
        ivp4_data = netbox.get(f"ipam/prefixes/{network_data.subnet_v4_id}/available-ips")
        ipv4 = ipaddress.IPv4Interface(ivp4_data[0]["address"])
        host_bits = int(str(int(ipv4.ip - int(ipv4.network.network_address))), 16)
        ipv6_address = network_data.subnet_v6.network_address + host_bits
        ipv6 = ipaddress.IPv6Interface((ipv6_address, network_data.subnet_v6.prefixlen))
        return str(ipv4), str(ipv6)

    def _update_dns(self):
        dnsgen = self.clippy.remote.query("P{P:dns::dnsgen}", single=True)
        self.logger.info("Updating DNS data")
        dnsgen.execute(command="update-dns", show_output=True)

    def _create_on_netbox(
        self, netbox: Netbox, hostname: str, virt_cluster: VirtCluster, hv: Hypervisor, network_data: NetworkData
    ) -> VirtualMachine:
        self.logger.info("Creating new VM %s on Netbox", hostname)
        vm = netbox.create_object(
            "virtualization/virtual-machines",
            {
                "name": hostname,
                "status": "staged",
                "cluster": virt_cluster.netbox_id,
                "device": hv.netbox_id,
                "platform": netbox.get_platform_id(self.args.os),
                "vcpus": self.args.cpu,
                "memory": int(self.args.mem * 1024),
                "disk": int(self.args.disk_space * 1000),
                "custom_fields": {
                    "disk_partition_layout": str(self.args.disk_layout),
                },
            },
        )

        self.logger.info("Creating primary interface")
        interface = netbox.create_object(
            "virtualization/interfaces",
            {
                "virtual_machine": vm["id"],
                "name": PRIMARY_INTERFACE_NAME,
                "mode": "access",
                "untagged_vlan": network_data.vlan_id,
            },
        )

        self.logger.info("Allocating IP addresses")
        ipv4, ipv6 = self._allocate_addresses(netbox, network_data)

        netbox.create_object(
            "ipam/ip-addresses",
            {
                "address": ipv4,
                "dns_name": self.args.fqdn,
                "assigned_object_type": "virtualization.vminterface",
                "assigned_object_id": interface["id"],
            },
        )

        netbox.create_object(
            "ipam/ip-addresses",
            {
                "address": ipv6,
                "dns_name": self.args.fqdn,
                "assigned_object_type": "virtualization.vminterface",
                "assigned_object_id": interface["id"],
            },
        )

        netbox.update_object(
            f"virtualization/virtual-machines/{vm['id']}",
            {
                "primary_ip4": {"address": ipv4},
                "primary_ip6": {"address": ipv6},
            },
        )

        return VirtualMachine(
            netbox_id=vm["id"],
            fqdn=self.args.fqdn,
            memory=self.args.mem,
            vcpus=self.args.cpu,
            disk=self.args.disk_space,
            network_interface_id=interface["id"],
            network_name=network_data.netbox_name,
        )

    def _wait_for_installation_to_complete(self, cluster: VirtCluster, hypervisor: Hypervisor, vm: VirtualMachine):
        for i in range(180):
            if cluster.is_running(hypervisor, vm):
                self.logger.debug("Installation is still running, retrying in 10s...")
                time.sleep(10)
                continue
            return
        raise RuntimeError("Timed out waiting for installation to complete")

    def _wait_for_ssh(self, remote: RemoteHosts):
        for i in range(24):
            try:
                remote.execute(command="true")
            except RemoteCommandError:
                self.logger.debug("Host is not unreachable, retrying in 5s...", exc_info=True)
                time.sleep(5)
                continue
            return
        raise RuntimeError("Timed out waiting for host to start")

    def _run_command_until_successful(self, command: str, remote: RemoteHosts):
        while True:
            try:
                remote.execute(command=command, show_output=True)
                break
            except RemoteCommandError as e:
                self.logger.warning("Failed to run '%s' on %s", command, self.args.fqdn)
                ans = input("> Retry [Y/n]? ")
                if ans.strip().lower() == "n":
                    raise e

    def execute(self) -> int:
        netbox = Netbox.from_config(self.clippy.config("netbox"))
        logging.getLogger("rover").setLevel(logging.getLogger("clippy").getEffectiveLevel())

        if not self.args.fqdn.endswith(".srv.majava.org"):
            raise ValueError("Host domain not supported")
        hostname = self.args.fqdn.split(".")[0]

        self.logger.debug("Querying Netbox for hypervisor %s", self.args.virt)
        virt_cluster, hypervisor = find_hv_cluster(netbox, self.clippy.remote, self.args.virt)
        network = self._get_network_data(netbox)

        vm = self._create_on_netbox(netbox, hostname, virt_cluster, hypervisor, network)
        self.logger.info("Netbox VM #%s created", vm.netbox_id)

        self._update_dns()

        self.logger.info("Creating VM on node %s in cluster %s", hypervisor.fqdn, virt_cluster.label)
        virt_cluster.create_vm(hypervisor=hypervisor, vm=vm)

        installer_host = self.clippy.remote_instance("installer").query(f"D{{{self.args.fqdn}}}")

        self.logger.info("Waiting for OS installation to complete..")
        self._wait_for_installation_to_complete(virt_cluster, hypervisor, vm)
        virt_cluster.start_vm(hypervisor, vm)
        self._wait_for_ssh(installer_host)
        self.logger.info("OS installation is complete")

        self.logger.info("Enrolling host in Puppet")
        puppet_host = self.clippy.puppet_hosts(self.args.fqdn)
        puppet_host.sign_certificates()
        self._run_command_until_successful(command="puppet agent -tv", remote=installer_host)

        self.logger.info("Rebooting to pick up networking changes")
        host = self.clippy.remote_hosts(f"D{{{self.args.fqdn}}}")
        host.reboot()

        return 0
