import time
from abc import ABC, abstractmethod
from dataclasses import dataclass

from clippy.exceptions import AutomationError
from clippy.remote import Remote, RemoteHosts

from rover.netbox import Netbox


class VirtError(AutomationError):
    pass


class VirtCluster(ABC):
    netbox_id: int
    label: str

    @abstractmethod
    def create_vm(self, hypervisor: "Hypervisor", vm: "VirtualMachine") -> None:
        pass

    @abstractmethod
    def is_running(self, hypervisor: "Hypervisor", vm: "VirtualMachine") -> bool:
        pass

    @abstractmethod
    def start_vm(self, hypervisor: "Hypervisor", vm: "VirtualMachine") -> None:
        pass

    @abstractmethod
    def disable_vm(self, hypervisor: "Hypervisor", fqdn: str) -> None:
        pass

    @abstractmethod
    def reset_vm(self, hypervisor: "Hypervisor", fqdn: str) -> None:
        pass


@dataclass
class Hypervisor:
    netbox_id: int
    fqdn: str


@dataclass
class VirtualMachine:
    netbox_id: int
    fqdn: str

    memory: int
    vcpus: int
    disk: int

    network_interface_id: int
    network_name: str


class ProxmoxCluster(VirtCluster):
    def __init__(
        self,
        remote: Remote,
        netbox_id: int,
        label: str,
    ):
        self._remote = remote
        self.netbox_id = netbox_id
        self.label = label

        self._id_cache: dict[tuple[str, str], int] = {}

    def _get_vm_id(self, hypervisor: "Hypervisor", vm: "VirtualMachine | str") -> int:
        fqdn = vm if isinstance(vm, str) else vm.fqdn
        cache_key = (hypervisor.fqdn, fqdn)
        if cache_key in self._id_cache:
            return self._id_cache[cache_key]

        hostname = fqdn.split(".")[0]
        result = (
            self._remote.query(f"D{{{hypervisor.fqdn}}}")
            .execute(command=f"qm list | grep {hostname} | awk '{{print $1}}'")
            .results
        )

        for nodeset, output in result:
            vm_id = int(output.message().decode().strip().split()[0])
            self._id_cache[cache_key] = vm_id
            return vm_id

        raise VirtError(f"Unable to determine virtual machine ID for {vm.fqdn}")

    def is_running(self, hypervisor: "Hypervisor", vm: "VirtualMachine") -> bool:
        vm_id = self._get_vm_id(hypervisor, vm)
        result = self._remote.query(f"D{{{hypervisor.fqdn}}}").execute(command=f"qm status {vm_id}").results
        for nodeset, output in result:
            return "running" in output.message().decode()
        raise VirtError(f"Unable to determine virtual machine status for {vm.fqdn}")

    def create_vm(self, hypervisor: Hypervisor, vm: VirtualMachine) -> None:
        self._remote.query(f"D{{{hypervisor.fqdn}}}").execute(
            command=f"/usr/local/sbin/majava-create-netbox-vm {vm.netbox_id}",
            show_output=True,
        )

    def start_vm(self, hypervisor: "Hypervisor", vm: "VirtualMachine") -> None:
        vm_id = self._get_vm_id(hypervisor, vm)
        self._remote.query(f"D{{{hypervisor.fqdn}}}").execute(command=f"qm start {vm_id}")
        time.sleep(10)

    def disable_vm(self, hypervisor: "Hypervisor", fqdn: str) -> None:
        vm_id = self._get_vm_id(hypervisor, fqdn)
        self._remote.query(f"D{{{hypervisor.fqdn}}}").execute(command=f"qm set {vm_id} --onboot false")

    def reset_vm(self, hypervisor: "Hypervisor", fqdn: str) -> None:
        vm_id = self._get_vm_id(hypervisor, fqdn)
        self._remote.query(f"D{{{hypervisor.fqdn}}}").execute(command=f"qm reset {vm_id}")


class GanetiCluster(VirtCluster):
    def __init__(
        self,
        remote: Remote,
        netbox: Netbox,
        netbox_id: int,
        label: str,
    ):
        self._remote = remote
        self._netbox = netbox
        self.netbox_id = netbox_id
        self.label = label

    def _get_mac(self, remote: RemoteHosts, vm: "VirtualMachine") -> str:
        """get the mac address for a specific VM"""
        for nodeset, output in remote.execute(command=f"gnt-instance list -o mac {vm.fqdn}").results:
            return output.message().decode().split("\n")[1].strip()
        raise VirtError(f"Unable to determine MAC address for {vm.fqdn}")

    def create_vm(self, hypervisor: "Hypervisor", vm: "VirtualMachine") -> None:
        remote = self._remote.query(f"D{{{hypervisor.fqdn}}}")
        remote.execute(
            command=(
                f"gnt-instance add -t plain -B minmem={vm.memory}G,maxmem={vm.memory}G,vcpus={vm.vcpus} "
                f"--net 0:network={vm.network_name},mac=generate --disk 0:size={vm.disk}g "
                f"-n {hypervisor.fqdn} --os-type noop --hypervisor-parameters kvm:boot_order=network --no-start "
                f"{vm.fqdn}"
            ),
            show_output=True,
        )

        mac_address = self._netbox.create_object(
            "dcim/mac-addresses",
            {
                "mac_address": self._get_mac(remote, vm),
                "assigned_object_type": "virtualization.vminterface",
                "assigned_object_id": vm.network_interface_id,
            },
        )

        self._netbox.update_object(
            f"virtualization/interfaces/{vm.network_interface_id}",
            {
                "primary_mac_address": mac_address["id"],
            },
        )
        self._netbox.update_object(f"virtualization/virtual-machines/{vm.netbox_id}", {"state": "active"})

        remote.execute(command=f"gnt-instance start {vm.fqdn}", show_output=True)

    def is_running(self, hypervisor: "Hypervisor", vm: "VirtualMachine") -> bool:
        result = (
            self._remote.query(f"D{{{hypervisor.fqdn}}}")
            .execute(command=f"gnt-instance list -o status {vm.fqdn}")
            .results
        )
        for nodeset, output in result:
            return "running" in output.message().decode()
        raise VirtError(f"Unable to determine virtual machine status for {vm.fqdn}")

    def start_vm(self, hypervisor: "Hypervisor", vm: "VirtualMachine") -> None:
        remote = self._remote.query(f"D{{{hypervisor.fqdn}}}")
        remote.execute(command=f"gnt-instance modify --hypervisor-parameters=boot_order=disk {vm.fqdn}")
        remote.execute(command=f"gnt-instance start {vm.fqdn}", show_output=True)

    def disable_vm(self, hypervisor: "Hypervisor", fqdn: str) -> None:
        self._remote.query(f"D{{{hypervisor.fqdn}}}").execute(command=f"gnt-instance shutdown {fqdn}")

    def reset_vm(self, hypervisor: "Hypervisor", fqdn: str) -> None:
        self._remote.query(f"D{{{hypervisor.fqdn}}}").execute(
            command=f"gnt-instance reboot --shutdown-timeout=1 {fqdn}"
        )


FIND_VIRT_NODE_QUERY = """
query get_node($host: String!) {
    device_list(filters: {name: {exact: $host}}) {
        id

        primary_ip4 {
            dns_name
        }

        cluster {
            id
            name

            type {
                name
            }
        }
    }
}
"""


FIND_VM_HYPERVISOR_QUERY = """
query get_hypervisor($name: String!) {
    virtual_machine_list(filters: {name: {exact: $name}}) {
        device {
            name
        }
    }
}
"""


def find_vm_hypervisor(netbox: Netbox, fqdn: str) -> str:
    nb_results = netbox.query(gql=FIND_VM_HYPERVISOR_QUERY, op="get_hypervisor", variables={"name": fqdn.split(".")[0]})
    nb_vm = nb_results["virtual_machine_list"][0]
    return nb_vm["device"]["name"]


def find_hv_cluster(netbox: Netbox, remote: Remote, fqdn: str) -> tuple["VirtCluster", "Hypervisor"]:
    nb_results = netbox.query(gql=FIND_VIRT_NODE_QUERY, op="get_node", variables={"host": fqdn})

    nb_virt = nb_results["device_list"][0]

    nb_id = int(nb_virt["id"])
    nb_fqdn = nb_virt["primary_ip4"]["dns_name"]

    cluster_type = nb_virt["cluster"]["type"]["name"]

    if cluster_type == "Proxmox VE":
        cluster = ProxmoxCluster(
            remote=remote,
            netbox_id=int(nb_virt["cluster"]["id"]),
            label=nb_virt["cluster"]["name"],
        )
    elif cluster_type == "Ganeti":
        cluster = GanetiCluster(
            remote=remote,
            netbox=netbox,
            netbox_id=int(nb_virt["cluster"]["id"]),
            label=nb_virt["cluster"]["name"],
        )
    else:
        raise AutomationError(f"Unknown cluster type '{cluster_type}'!")

    hv = Hypervisor(
        netbox_id=nb_id,
        fqdn=nb_fqdn,
    )

    return cluster, hv
